/*================================================================================================================================
# TriangleScanLine.cpp	: 
# Date		            : 30 July  1998 ; based on QuadScanLine.cpp
# Date		            : 10 Aptil 2002
# Date                  : 02 July  2002 
#
# Author	: A.Sarafopoulos
# Purpose: 
#		Triangle scanline algorithm
=================================================================================================================================*/

/*
 * Include files
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sstream>

using namespace std;

struct point {
	int x,y;
	point() { x=y=0; }
	point(int x, int y) { point::x=x; point::y=y; }
	point(const point& p) 	{ x = p.x; y = p.y; }
	~point() { ; }

	point& operator=(const point& p) { x = p.x; y = p.y; return ( *this ); }

	friend bool operator<(point& p1, point& p2) { 
		if ( p1.y == p2.y )
			return( p1.x < p2.x )? true : false;	
		else
			return( p1.y < p2.y )? true : false;
	}

	friend bool operator>(point& p1, point& p2) { 
		if (p1.y == p2.y)
			return( p1.x > p2.x )? true : false;	
		else
			return( p1.y > p2.y )? true : false;
	}
};


struct edge_t {
	point a, b;
	int deltax, deltay;
	int xmin, ymax;
	int increment;

	void set_increment() {
		if (deltax >= 0) {
		   //  slope in [0, 1) else slope in [1, +infinity )
			increment = deltax >= deltay ? 0 : deltay;
		} else {
			// slope in [-1, 0] else slope in (-infinity, -1]
			increment = abs(deltax) >= deltay ? -deltax : 0;
		}
	}

	edge_t(): a(), b() { deltax=deltay=0; xmin=ymax=0; increment=0; }
	edge_t(const edge_t& e) 
		: a(e.a), b(e.b), deltax(e.deltax), deltay(e.deltay), xmin(e.xmin), ymax(e.ymax)
	{ increment=e.increment; }
		
	edge_t(point bottom, point top) {
		a=bottom;
		b=top;
		deltax = b.x-a.x;
		deltay = b.y-a.y;
		xmin = a.x; // error->		(a.x < b.x) ? a.x : b.x; !
		ymax = b.y; // redundant->	(a.y > b.y) ? a.y : b.y; !

		if ( deltax >= 0 ) {
			increment = (deltax >= deltay) ? 0 : deltay;
		} else {
			increment = (abs(deltax) >= deltay) ? deltay : 0;  //-deltax : 0;
		}
	}

	~edge_t() { ; }

	edge_t& operator=(const edge_t& e) {
		a=e.a; 
		b=e.b;
		deltax=e.deltax; 
		deltay=e.deltay;
		xmin=e.xmin; 
		ymax=e.ymax;
		increment=e.increment;
		return ( *this );
	}
};

/*
 * Static function prototypes
 */
static int scanconvert_left_edge(edge_t& e);
static int scanconvert_right_edge(edge_t& e);
static void TriangleScanline(point *varray);
static void QuadScanline(point *varray);

/*
 * Static variables
 */
static void (*draw_pixel_callback) (int, int, bool, void*);


/*******************************************************************************************
 *
 * SetDrawPixelCallBack
 *
 ******************************************************************************************/
void Set_draw_pixel_callback(void (*func)(int, int, bool, void*)) {
	draw_pixel_callback = func;
}

/******************************************************************************************** 
 * 
 * scan_convert_triangle_quad:
 *  The Scanline class, stores internally data that 
 *  represent two polygons, each polygon is either a 
 *  triangle or a quad. The data is stored using the following format:
 *   
 *   array[0] -> holds the number of vertecies stored in the array, which is either 3, or 4.
 *               If the value stored is other than 3 or 4 an error has occured!
 *   array[1], array[2] -> hold the (x,y) co-ordinates of vertex 1
 *   array[3], array[4] -> hold the (x,y) co-ordinates of vertex 2
 *   array[5], array[6] -> hold the (x,y) co-ordinates of vertex 3
 *   array[7], array[8] -> hold the (x,y) co-ordinates of vertex 4
 *
 ********************************************************************************************/
void Scan_convert_triangle_quad(int *array) {
	if (array[0] == 3) {
	   // this is a triangle
	   // first convert convert the data array to an array of points
		point triangle_alpha[3];
		triangle_alpha[0] = point(array[1], array[2]);
		triangle_alpha[1] = point(array[3], array[4]);
		triangle_alpha[2] = point(array[5], array[6]);
		// scan-convert
		TriangleScanline(triangle_alpha);
	} else if (array[0] == 4) {
	   // this is a quad
	   // first convert convert the data array to an array of points
		point quad_alpha[4];
		quad_alpha[0] = point(array[1], array[2]);
		quad_alpha[1] = point(array[3], array[4]);
		quad_alpha[2] = point(array[5], array[6]);
		quad_alpha[3] = point(array[7], array[8]);
		// scan-convert
		QuadScanline(quad_alpha);
	}
}


/*******************************************************************************************
 *
 * bouble_point : bouble sort 'n' points in array 'x'
 *
 ******************************************************************************************/
static inline void bouble_point(point x[], int n) {
	point hold;
	int j, pass;
	bool switched= true;
	for ( pass=0; ( pass < n-1 ) && ( switched == true ); pass++ ) {
		// outer loop controls the number of passes
		// initially no interchanges have been made
		switched = false;
		for ( j=0; j < n-pass-1; j++ ) {
			if ( x[j] > x[j+1] ) {
				// elements out of order interchange is necessary
				switched = true;
				hold = x[j];
				x[j] = x[j+1];
				x[j+1] = hold;
			}
		}
	}
}

/*******************************************************************************************
 *
 * SWAP:
 * 
 ******************************************************************************************/
static inline void SWAP( edge_t& A, edge_t& B) {	
	edge_t e_temp;
	e_temp=A; A=B; B=e_temp; 
}

/*******************************************************************************************
 *
 * order_the_edges : 
 *
 ******************************************************************************************/
static inline void order_the_edges(edge_t& A, edge_t& B) {
	if ( A.deltax==0 && B.deltax<0 ) {
		SWAP(A, B);
		return ;
	}
	if ( B.deltax==0 && A.deltax>0 ) {
		SWAP(A, B);
		return ;
	}

	float slopeA = ((float)A.deltay)/((float)A.deltax);
	float slopeB = ((float)B.deltay)/((float)B.deltax);
	
	if ( slopeA>=0 && slopeB>=0 ) {
		if ( slopeA<slopeB ) SWAP(A, B);
	} else if ( slopeA<0 && slopeB<0 ) {
		if ( slopeA<slopeB ) SWAP(A, B);
	} else if ( slopeB<0 ) {
		SWAP(A, B);
	} else
		; // error()
}

/********************************************************************************************
 * 
 * drawscanline : 
 * 
 *******************************************************************************************/
static inline void drawscanline(int left_x, int right_x, int y) {
	//draw_pixel_callback(left_x, y, true, NULL);
	//left_x++;
	int v_left = left_x;
	while (left_x < right_x) {
		if (v_left == left_x) {
			draw_pixel_callback(left_x, y, true, NULL);
		} else {
			draw_pixel_callback(left_x, y, false, NULL);
		}
		left_x++;
	}
	if (left_x == right_x) {
		draw_pixel_callback(left_x, y, true, NULL);
	}
}

/********************************************************************************************
 *
 * scanconvert_left_edge :
 *
 *******************************************************************************************/
static int scanconvert_left_edge( edge_t& e ) {
	if ( e.deltax >= 0 ) {
		if ( e.deltax >= e.deltay ) {	// slope in [0, 1)
			while( 1 ) {
				e.increment += e.deltay;
				if ( e.increment >= e.deltax ) {
					e.xmin++;
					e.increment-=e.deltax;
					break;
				} else {
					e.xmin++;
				}
			}
		} else {	// slope in [1, +infinity )
			e.increment += e.deltax;
			if ( e.increment  >  e.deltay )	{
				e.xmin++;				
				e.increment-=e.deltay;	
			}
		}
	} else {
		if (  abs(e.deltax) >= e.deltay ) {	// slope in [-1, 0]
			while ( 1 ) {	
				e.increment += e.deltay;
				if ( e.increment > -e.deltax ) {   
					e.xmin--;
					e.increment -= -e.deltax;
					break;
				} else {
					e.xmin--;
				}
			}
		} else	{	// slope in (-infinity, -1]
			e.increment += e.deltax;
			if ( abs(e.increment) >= e.deltay )
			{
				e.xmin--;
				e.increment +=e.deltay;
			}
		}
	}

	return ( e.xmin );
}

/********************************************************************************************
 * 
 * scanconvert_right_edge : 
 * 
 *******************************************************************************************/
static int scanconvert_right_edge( edge_t& e ) {
	int xmin = e.xmin;
	if ( e.deltax >= 0 ) {
		if ( e.deltax >= e.deltay ) {	// slope in [0, 1)
			while( 1 ) {
				e.increment += e.deltay;
				if ( e.increment >= e.deltax ) {
					e.xmin++;
					e.increment-=e.deltax;
					break;
				} else {
					e.xmin++;
				}
			}
		} else {	// slope in [1, +infinity )
			e.increment += e.deltax;
			if ( e.increment  >  e.deltay )	{
				e.xmin++;				
				e.increment-=e.deltay;	
			}
		}
	} else {
		if (  abs(e.deltax) >= e.deltay ) {	// slope in [-1, 0]
			while ( 1 ) {	
				e.increment += e.deltay;
				if ( e.increment > -e.deltax ) {   
					e.xmin--;
					e.increment -= -e.deltax;
					break;
				} else {
					e.xmin--;
				}
			}
		} else {	// slope in (-infinity, -1]
			e.increment += e.deltax;
			if ( abs(e.increment) >= e.deltay ) {
				e.xmin--;
				e.increment +=e.deltay;
			}
		}
	}

	return ( e.xmin-1 );
}



/********************************************************************************************
 * 
 * TriangleScanline :
 *
 *******************************************************************************************/
static void TriangleScanline(point *varray) {
	int minL, minR;
	// sort the quadrilateral's vertecies according to their y-coordinate value
	int top=2, middle=1, bottom=0;
	bouble_point(varray, 3);
	
	// Generate Edges to be scanned 
	edge_t A(varray[bottom], varray[middle]);
	edge_t B(varray[bottom], varray[top]);
	
	// correctly order the Edges 
	order_the_edges(A, B);
	
	// in case the top or bottom scanlines are horizontal, (slope == 0)
	int i=0;
	if (varray[bottom].y == varray[middle].y) {
		// horizontal bottom scanline
		drawscanline(varray[bottom].x, varray[middle].x-1, varray[bottom].y);
		//drawscanline(varray[bottom].x, varray[middle].x, varray[bottom].y);
		i=2;
	} else if (varray[middle].y == varray[top].y) {  
		// horizontal top scanline
		i=1;
	} else {
		i=2;
	}
	
	// set Y counter to bottom scanline
	int Y =varray[bottom].y ; 
	
	// generate scanlines
	minL = A.xmin; 
	minR = B.xmin;
	while ( i>0 ) {
		while ( Y < A.ymax  &&   Y < B.ymax ) {
			if (Y == varray[bottom].y) {
				//drawscanline( minL+1, minR, Y );
				drawscanline( minL, minR, Y );
			} else {
				drawscanline( minL, minR, Y );
			}
			minL = scanconvert_left_edge( A );
			minR = scanconvert_right_edge( B );
			Y++;
		}
		if ( A.ymax == Y )A = edge_t(varray[middle], varray[top]);
		if ( B.ymax == Y )B = edge_t(varray[middle], varray[top]);
		i--;
	} // end for loop 
}

/*******************************************************************************************
 *
 * SWAP:
 * 
 ******************************************************************************************/
static inline void SWAP( edge_t& A, edge_t& B, int& a, int& b ) {	
	int i_temp;
	i_temp=a; a=b; b=i_temp;
	edge_t e_temp;
	e_temp=A; A=B; B=e_temp; 
}

/*******************************************************************************************
 *
 * order_the_edges:
 * 
 ******************************************************************************************/
static inline void order_the_edges( edge_t& A, edge_t& B, int& a, int& b ) {
	if ( A.deltax==0 && B.deltax<0 ) {
		SWAP(A, B, a, b);
		return ;
	}
	if ( B.deltax==0 && A.deltax>0 ) {
		SWAP(A, B, a, b);
		return ;
	}

	float slopeA = ((float)A.deltay)/((float)A.deltax);
	float slopeB = ((float)B.deltay)/((float)B.deltax);

	if ( slopeA>=0 && slopeB>=0 ) {
		if ( slopeA<slopeB ) SWAP(A, B, a, b);
	} else if ( slopeA<0 && slopeB<0 ) {
		if ( slopeA<slopeB ) SWAP(A, B, a, b);
	} else if ( slopeB<0 ) {
		 SWAP(A, B, a, b);
	} else
		; // error()
}

/******************************************************************************************** 
 * 
 * QuadScanline :
 * 
 ********************************************************************************************/
static void QuadScanline(point *varray) {
	int minL, minR;
	//
	// sort the quadrilateral's vertecies according to their y-coordinate value
	//
	int top=3, b=2, a=1, bottom=0;
	bouble_point(varray, 4);

	//
	// Generate Edges to be scanned 
	// 
	edge_t A( varray[bottom], varray[a] );
	edge_t B( varray[bottom], varray[b] );


	//
	// in case the top or bottom scanlines are horizontal, (slope == 0)
	//
	int i=0;
	if ( varray[top].y == varray[b].y ) {  
		 // horizontal top scanline
		i=2;
	} else if ( varray[bottom].y == varray[a].y ) {  
		 // horizontal bottom scanline
		i=2;
		drawscanline( varray[bottom].x, varray[a].x, varray[bottom].y );
	} else {
		i=3;
	}

	//
	// correctly order the Edges 
	//
	order_the_edges(A, B, a, b);

	// 
	//  set Y counter to bottom scanline
	//
	int Y =varray[bottom].y ; 

	//
	// scanline!
	//
	minL = A.xmin; 
	minR = B.xmin;
	while ( i>0 ) {
	  while ( Y < A.ymax  &&   Y < B.ymax ) {
			drawscanline( minL, minR, Y );
			minL = scanconvert_left_edge( A );
			minR = scanconvert_right_edge( B );
			Y++;
		} 

		if ( A.ymax == Y )A = edge_t( varray[a], varray[top] );
		if ( B.ymax == Y )B = edge_t( varray[b], varray[top] );
		i--;
	} // end for loop 

}


