/*==========================================================================================
# Scanline_01.cpp
# Date   : 01 July 2002
# Author	: Anargyros Sarafopoulos
===========================================================================================*/

/*
 * Include
 */
#include "scanline.h"
#include "TriangleScanLine.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdlib.h>
#include <fstream>

using namespace std;

static ofstream err_out("scanline_errors.txt", ios::out);


/*
 * draw_a_line :
 */
static inline void draw_a_line( int x0, int y0, int x1, int y1 ) {
	glBegin( GL_LINES );
		glVertex2i(x0+GRID_DX/2, y0+GRID_DY/2);
		glVertex2i(x1+GRID_DX/2, y1+GRID_DY/2);
	glEnd();
}

/*
 * Draws the outline of a polygon stored in "array".
 * The polygon data is stored using the following format:
 *   
 *   array[0] -> holds the number "n" of vertecies stored in the array.
 *   array[1], array[2] -> hold the (x,y) co-ordinates of vertex 1
 *   array[3], array[4] -> hold the (x,y) co-ordinates of vertex 2
 *   array[5], array[6] -> hold the (x,y) co-ordinates of vertex 3
 *   ...
 *   array[n*2-1], array[n*2] -> holds the (x,y) co-ordinates of the n-th vertex
 *
 */
static void draw_polygon_outline(int *array){
  for (int p=0, i=0; p<array[0]; p++,i+=2) {
	 if (p<array[0]-1) {
		draw_a_line(array[i+1]*GRID_DX, array[i+2]*GRID_DY, 
						array[i+3]*GRID_DX, array[i+4]*GRID_DY);
	 } else {
		draw_a_line(array[i+1]*GRID_DX, array[i+2]*GRID_DY, 
						array[0+1]*GRID_DX, array[0+2]*GRID_DY);
	 }
  }
}




/* 
 * plot_pixel : draw a grid cell, and save it into 'pixels'
 */
static inline void plot_pixel(int left, int bottom,  int right, int top) {
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glRecti( left, bottom,  right, top );
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
}


static float in_R =0.0f, in_G =0.0f, in_B =0.0f, in_A =1.0f;
static float out_R=0.3f, out_G=0.8f, out_B=0.3f, out_A=1.0f;

static void set_in_pixel_colour(float r, float g, float b, float a=1.0f) {
	//glColor4f(in_R, in_G, in_B, in_A);
	in_R = r;
	in_G = g;
	in_B = b;
	in_A = a;
}

static void set_out_pixel_colour(float r, float g, float b, float a=1.0f) {
	//glColor4f(out_R, out_G, out_B, out_A);
	out_R = r; 
	out_G = g; 
	out_B = b; 
	out_A = a;
}

/* 
 * draw_pixel_callback
 * Input: x,y : left-bottom pixel co-ordinate
 *          v : true if the pixel with (x,y) co-ordinates is the 
 *              start or end point of the scanline, false otherwise
 *         ptr: data
 */
static void draw_pixel_callback(int x, int y, bool v, void* ptr) {
	ptr=NULL;
	x *= GRID_DX;
	y *= GRID_DY;
	if (v==true) {
		glColor4f(out_R, out_G, out_B, out_A);
		plot_pixel( x, y, x+GRID_DX, y+GRID_DY );
	} else {
		glColor4f(in_R, in_G, in_B, in_A);
		plot_pixel( x, y, x+GRID_DX, y+GRID_DY );
	}
}


/*
 * Display Call-back 
 */
void Scanline::Display() {
	glClearColor(1.0, 1.0, 1.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	// draw a grid of quads
	int grid_width  = GetWidth()/GRID_DX;
	int grid_height = GetHeight()/GRID_DY;
	for (int i=0; i<grid_height+1; i++) {
		for (int j=0; j<grid_width+1; j++) {
			glColor3f( 0.0f, 0.0f, 0.0f );
			//int left  = j*GRID_DX,     bottom = i*GRID_DY;
			//int right = (j+1)*GRID_DX, top    = (i+1)*GRID_DY;
			//glRecti(left, bottom, right, top);
			glRecti(j*GRID_DX , i*GRID_DY, (j+1)*GRID_DX , (i+1)*GRID_DY);
		}
	}
	
	//
	// draw polygon outlines
	//
	glColor3f(1.0f, 0.0f, 0.0f);
	draw_polygon_outline(plg_01);
	draw_polygon_outline(plg_02);
	
	//
	// draw polygon scalines
	//
	set_out_pixel_colour(1, 0, 0);
	Scan_convert_triangle_quad(plg_01);
	
	set_out_pixel_colour(0, 0, 1);
	Scan_convert_triangle_quad(plg_02);

	SwapBuffers();
}

/*
 * Reshape Call-back
 */
void Scanline::Reshape () {
	int w = GetWidth();
	int h = GetHeight();
	glViewport( 0, 0, (GLsizei) w, (GLsizei) h );
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D( 0.0, (GLdouble) w, 0.0, (GLdouble) h );
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/*
 * update: idle-function call-back
 */
void Scanline::Timer() { 
	PostRedisplay();
}


/*
 * init OpenGL
 */
void Scanline::Init() {
	Set_draw_pixel_callback(draw_pixel_callback);
	
	glEnable(GL_DEPTH_TEST);    // allow z-buffer 
	glShadeModel(GL_FLAT);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glEnable(GL_COLOR_LOGIC_OP);
	glEnable(GL_BLEND);

	int red=0;
	glGetIntegerv(GL_RED_BITS, &red);
	cout<<"GL_RED_BITS   "<<red<<endl;
	//err_out<<"GL_RED_BITS   "<<red<<endl;
	
	int green=0;
	glGetIntegerv(GL_GREEN_BITS, &green);
	cout<<"GL_GREEN_BITS "<<green<<endl;
	//err_out<<"GL_GREEN_BITS "<<green<<endl;
	
	int blue=0;
	glGetIntegerv(GL_BLUE_BITS, &blue);
	cout<<"GL_BLUE_BITS  "<<blue<<endl;
	//err_out<<"GL_BLUE_BITS  "<<blue<<endl;
	
	int alpha=0;
	glGetIntegerv(GL_ALPHA_BITS, &alpha);
	cout<<"GL_ALPHA_BITS "<<alpha<<endl;
	//err_out<<"GL_ALPHA_BITS "<<alpha<<endl;
	
	int accum_red=0;
	glGetIntegerv(GL_ACCUM_RED_BITS, &accum_red);
	cout<<"GL_ACCUM_RED_BITS   "<<accum_red<<endl;
	//err_out<<"GL_ACCUM_RED_BITS   "<<accum_red<<endl;
	
	int accum_green=0;
	glGetIntegerv(GL_ACCUM_GREEN_BITS, &accum_green);
	cout<<"GL_ACCUM_GREEN_BITS "<<accum_green<<endl;
	//err_out<<"GL_ACCUM_GREEN_BITS "<<accum_green<<endl;
	
	int accum_blue=0;
	glGetIntegerv(GL_ACCUM_BLUE_BITS, &accum_blue);
	cout<<"GL_ACCUM_BLUE_BITS  "<<accum_blue<<endl;
	//err_out<<"GL_ACCUM_BLUE_BITS  "<<accum_blue<<endl;

	int accum_alpha=0;
	glGetIntegerv(GL_ACCUM_ALPHA_BITS, &accum_alpha);
	cout<<"GL_ACCUM_ALPHA_BITS "<<accum_alpha<<endl;
	//err_out<<"GL_ACCUM_ALPHA_BITS "<<accum_alpha<<endl;

	int aux_buff=0;
	glGetIntegerv(GL_AUX_BUFFERS, &aux_buff);
	cout<<"GL_AUX_BUFFERS "<<aux_buff<<endl;
	//err_out<<"GL_AUX_BUFFERS "<<aux_buff<<endl;

	const GLubyte *glver =  glGetString(GL_VERSION);
	cout<<glver<<endl;
	//err_out<<glver<<endl;

	const GLubyte *gluver =  gluGetString(GLU_VERSION);
	cout<<gluver<<endl;
	//err_out<<gluver<<endl;

#ifdef GLU_VERSION_1_3
	const GLubyte *imagex  = glGetString(GL_EXTENSIONS);
	const GLubyte *extName = (GLubyte*)"GL_ARB_imaging";
	if (gluCheckExtension(extName, imagex) == GL_TRUE) {
		cout<<"GL_ARB_imaging"<<endl;
		//err_out<<"GL_ARB_imaging"<<endl;
	} else {
		cout<<"No support for the ARB Imaging extension"<<endl;
		//err_out<<"No support for the ARB Imaging extension"<<endl;
	}
#endif // GLU_VERSION_1_3

}

/*
 * Used to by Togl to create new instances of the "Scanline" widget
 */
ToglWidget* Scanline::Clone() { 
	return new Scanline(); 
}

/*
 * Init tcl/tk and togl
 */
extern "C" EXPORT(int, Scanline_Init)( Tcl_Interp *interp ) {
#ifdef USE_TCL_STUBS
  if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {return TCL_ERROR;}
#endif
#ifdef USE_TK_STUBS
  if (Tk_InitStubs(interp, "8.1", 0) == NULL) {return TCL_ERROR;}
#endif

   if (Togl_Init(interp) == TCL_ERROR) {
      return TCL_ERROR;
   }

#ifdef macintosh
	 Togl_MacSetupMainInterp(interp);
#endif

	// init Togl library callbacks
	ToglMaster::InitTogl();
	// register ToglWidget classes
	ToglMaster::RegisterWidgetFunc("Scanline", Scanline::Clone);

   return TCL_OK;
}

