//----------------------------------------------
//
// C++ interface "ToglMaster", for "Togl"
// Anargyros Sarafopoulos 08/June/01
// asarafop@bournemouth.ac.uk
//
//---------------------------------------------

#include "toglmaster.h"
#include <sstream>
#include <iostream>
#include <cstring>

using namespace std;

/*----------------------------------------------------------------------
 *
 *  Init static members of ToglMaster
 *
  ---------------------------------------------------------------------- */

ClassTable ToglMaster::toglClasses;  //Table of widget classes 
WidgetTable ToglMaster::toglWidgets; //Table of widgets

/*----------------------------------------------------------------------
 #
 # ToglCmd constructor
 #
  ---------------------------------------------------------------------- */
ToglCmd::ToglCmd(
		string str, // command name
		ToglWidget &w // widget that invokes "this" command
) { 
	name = str; // init command name
	w.RegisterCommand(name, this); // register command
}

/*----------------------------------------------------------------------
 #
 # ToglWidget::RegisterWidget, 
 # registers an active widget with ToglMaster
 # called by ToglMaster to insert "this" widget
 # into the ToglMaster::toglWidgets table
 #
  ---------------------------------------------------------------------- */
void ToglWidget::RegisterWidget(Togl *togl, string &str) {
	// init Togl pointer and widget id
	ToglWidget::togl = togl;
	ToglWidget::id   = str;
	// register "this" widget with ToglMaster
	ToglMaster::RegisterToglWidget(id.c_str(), *this);
	// register command strings used by this widget 
	CommandTable::iterator p;
	for(p=commands.begin(); p!=commands.end(); ++p) {
		ToglMaster::RegisterCmdStr(p->first.c_str());
	}
}

/*----------------------------------------------------------------------
 #
 # ToglWidget::RegisterCommand 
 # Called by a ToglCmd object in order to insert
 # itself (or another ToglCmd) in "this" ToglWidget::commands table
 #
  ---------------------------------------------------------------------- */
void ToglWidget::RegisterCommand(std::string cmdname, ToglCmd* cmd) {
	std::pair<std::string, ToglCmd*> p1(cmdname, cmd);
	commands.insert(p1);
}

/*----------------------------------------------------------------------
 #
 # ToglWidget::Cmd
 # invokes a widget command
 #
  ---------------------------------------------------------------------- */
bool ToglWidget::Cmd(int argc, const char *argv[]) {
	typedef CommandTable::const_iterator I;
	pair<I,I> v = commands.equal_range(argv[1]);
	for (I i=v.first; i != v.second; ++i) {
		try {
			if ((*(i->second)).Cmd(argc, argv) == true) {
				return (true);
			}
		}
		catch (ToglCmd::WrongArgs argm) {
			Tcl_Interp *interp = tkgl::GetInterp();
			Tcl_AppendResult(interp, (char *) argm.GetMsg(), NULL);
			return (false);
		}
	}
	return (false);
}

/*----------------------------------------------------------------------
 #
 # ToglMaster method that loads up the private and 
 # static associative array (key, value) pairs of :
 # "map<srting, ToglWidget*> toglWidgets" 
 #
 # Output:
 #   If a value already exists in the "toglWidgets" array with key
 #   equal to "wname" the operation fails, otherwise the operation succeeds.
 #   The method returns "true" if the operation succeeds, "false" otherwise
 #
 # On Exit :
 #   In the case where the operation succeeds, "widget" becomes a 
 #   value of the associative array "toglWidgets" with key set to "wname" 
 #
  ---------------------------------------------------------------------- */
bool ToglMaster::RegisterToglWidget(
		const char *wname, // widget's key,
		ToglWidget &widget // widget object to become value of "toglWidgets"
)  {
	pair<string, ToglWidget*> p1(wname, &widget);
	pair<WidgetTable::iterator, bool> p = toglWidgets.insert(p1);
	if (p.second) {
		; // OK
		return true;
	} else {
		cerr<<"Error in ToglMaster:RegisterToglWidget(char *, ToglWidget&); # ";
		cerr<<"the widget : "<<"\""<<wname<<"\" already exists!"<<endl;
		return false;
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster Init callback invoked by Togl. It accepts
 # a pointer to 'C' struct Togl (as all Togl callbacks do) used 
 # to retrieve the widget's -Ident string that comes in the form :
 # "classname::instanceName". 
 # Where the "className" part of the sting signifies the class of a 
 # widget to be created. A new widget of class "className" is created  
 # whose "key" in the toglWidgets associative array is specified by: 
 # "className::instanceN", where "N" is the number of instaces of class
 # "className" in the table.
 #
 # The class name has to be specifieet  in the widget's "-Ident" attribute.
 # The format of the "-Ident" string is : "classname::instanceName", where 
 # the first occurence of "::" separates "className" from "instanceName".
 # For exmaple if the widget's -Ident is "IFS::fern", the class name is
 # "IFS".
 #
  ---------------------------------------------------------------------- */

static string makeNewKey(
		string className, // widget class name
		WidgetTable &toglWidgets // table of active widgets
) {
	/*
	 * find how many instances of widgets of class 
	 * "className" are there in the "toglWidgets" table
	 */
	int counter = 1;
	WidgetTable::iterator p;
	for(p = toglWidgets.begin(); p !=toglWidgets.end(); ++p) {
		string to_compare = className + "::";
//#ifdef WIN32
		if (p->first.compare(0, className.length()+2, to_compare) == 0) {
			counter++;
		}
//#else
//		if (p->first.compare(to_compare, 0, className.length()+2) == 0) {
//			counter++;
//		}
//#endif // WIN32
	}
	
	/*
	 * we construct the key, based on the widget's class
	 * and the number of instaces of widgets of the
	 * same class in the "toglWidgets" table
	 */
	ostringstream key;
	key<<className<<"::instance"<<counter;
	return key.str();
}

void ToglMaster::Init(
		Togl *togl // pointer to Togl 'C' struct
) {
	char *ident = Togl_Ident(togl); // get Togl identifiction string
	if (ident != NULL) {
		// get widget's class name from -Ident string
		string className(ident);
		unsigned int v = className.find("::");
		if (v != string::npos) {
			className = className.substr(0, v);
		}
		CreateWidget ni = toglClasses[className];
		if (ni != NULL) {
			// OK, class "calssName" is registered 
			string key = makeNewKey(className, toglWidgets);
			WidgetTable::iterator p = toglWidgets.find(key);
			if (p == toglWidgets.end()) {
				ToglWidget *wpt = (*ni)(); // create new instace
				wpt->RegisterWidget(togl, key);
				Togl_SetClientData(togl, (ClientData)key.c_str());
				wpt->Init();
			} else {
				;// the widget already exists
			}
		} else {
			;// Oh... "className" is not registered!
		}
	} else {
		;// no class id, or widget name!
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster Reshape callback to be invoked by Togl. It accepts
 # a pointer to 'C' struct Togl (as all Togl callbacks do) used
 # to retrieve the widget's ClientData pointer. The pointer is the key to 
 # the value of a widget whose responsibility is to invoke "ToglWidget::Reshape"
 #
  ---------------------------------------------------------------------- */
void ToglMaster::Reshape(
		Togl *togl // pointer to Togl 'C' struct
) {
	char *key = (char *) Togl_GetClientData(togl); // get Togl identification string
	if (key != NULL) {
		WidgetTable::iterator p = toglWidgets.find(key);
		if (p != toglWidgets.end()) {
			ToglWidget *w = p->second;
			if (w != NULL) {
				w->Reshape();
			} else {
				;// no widget with this id!
			}
		}
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster Display callback to be invoked by Togl. It accepts
 # a pointer to 'C' struct Togl (as all Togl callbacks do) used
 # to retrieve the widget's ClientData pointer. The pointer is the key to 
 # the value of a widget whose responsibility is to invoke "ToglWidget::Display"
 #
  ---------------------------------------------------------------------- */
void ToglMaster::Display(
		Togl *togl // pointer to Togl 'C' struct
) {
	char *key = (char *) Togl_GetClientData(togl); // identification string
	if (key != NULL) {
		WidgetTable::iterator p = toglWidgets.find(key);
		if (p != toglWidgets.end()) {
			ToglWidget *w = p->second;
			if (w != NULL) {
				w->Display();
			} else {
				;// no widget with this id!
			}
		}
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster Destroy callback to be invoked by Togl. It accepts
 # a pointer to 'C' struct Togl (as all Togl callbacks do) used
 # to retrieve the widget's ClientData pointer. The pointer is the key to 
 # the value of a widget whose responsibility is to invoke "ToglWidget::Destroy"
 #
  ---------------------------------------------------------------------- */
void ToglMaster::Destroy(
		Togl *togl // pointer to Togl 'C' struct
) {
	char *key = (char *) Togl_GetClientData(togl); // get Togl identification string
	if (key != NULL) {
		WidgetTable::iterator p = toglWidgets.find(key);
		if (p != toglWidgets.end()) {
			ToglWidget *w = p->second;
			w->Destroy(); // destroy the widget's geometry 
			delete w;         // de-allocate object 
			p->second = NULL; // init widget value to NULL
			toglWidgets.erase(p);
		} else {
			;// no widget with this id!
		}
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster Timer callback to be invoked by Togl. It accepts
 # a pointer to 'C' struct Togl (as all Togl callbacks do), used
 # to retrieve the widget's ClientData pointer. The pointer is the key to 
 # the value of a widget whose responsibility is to invoke "ToglWidget::Timer"
 #
  ---------------------------------------------------------------------- */
void ToglMaster::Timer(
		Togl *togl // pointer to Togl 'C' struct
) {
	char *key = (char *) Togl_GetClientData(togl); // get Togl identification string
	if (key != NULL) {
		WidgetTable::iterator p = toglWidgets.find(key);
		if (p != toglWidgets.end()) {
			ToglWidget *w = p->second;
			if (w != NULL) {
				w->Timer();
			} else {
			;// no widget with this id!
			}
		}
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster Overlay callback to be invoked by Togl. It accepts
 # a pointer to 'C' struct Togl (as all Togl callbacks do) used
 # to retrieve the widget's ClientData pointer. The pointer is the key to 
 # the value of a widget whose responsibility is to invoke "ToglWidget::Overlay"
 #
  ---------------------------------------------------------------------- */
void ToglMaster::Overlay(
		Togl *togl // pointer to Togl 'C' struct
) {
	char *key = (char *) Togl_GetClientData(togl); // get Togl identification string
	if (key != NULL) {
		WidgetTable::iterator p = toglWidgets.find(key);
		if (p != toglWidgets.end()) {
			ToglWidget *w = p->second;
			if (w != NULL) {
				w->Overlay();
			} else {
				;// no widget with this id!
			}
		}
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster Cmd callback to be invoked by Togl. It accepts
 # a pointer to 'C' struct Togl (as all Togl callbacks do) used, here, in order
 # to retrieve the widget's ClientData pointer. The pointer is the key to 
 # the value of a widget whose responsibility is to evoke "ToglWidget::Cmd"
 # The other two arguments store 'C' style "argc" and "argv" command line 
 # input.
 #
  ---------------------------------------------------------------------- */
int ToglMaster::Togl_command(
		Togl *togl,    // pointer to Togl 'C' struct
		int argc,      // number of argument the command accepts 
		const char *argv[]   // array of points to the commands arguments
) {
	char *key = (char *) Togl_GetClientData(togl); // get Togl identification string
	if (key != NULL) {
		WidgetTable::iterator p = toglWidgets.find(key);
		if (p != toglWidgets.end()) {
			ToglWidget *w = p->second;
			if (w != NULL) {
				//
				// In case the command changes the OpenGL
				// state bind the OpenGL rendering context 
				// to the current Togl widget.
				//
				w->MakeCurrent();
				// execute the widget's command
				if (!w->Cmd(argc, argv))
					return TCL_ERROR;
			} else {
				;// no widget with this id!
			}
		}
	}
	return TCL_OK;
}


/*----------------------------------------------------------------------
 #
 # ToglMaster widget creation method: It loads up the 
 # private and static associative array (key, value) pairs of 
 # "map<srting, CreateWidget> toglClasses" 
 #
 # Output:
 #   If a value already exists in the "toglClasses" array with key
 #   equal to "name" the operation fails, otherwise the operation succeeds.
 #   The method returns "true" if the operation succeeds, "false" otherwise
 #
 # On Exit :
 #   In the case where the operation succeeds, on exit the "func" is a 
 #   value of the associative array "toglClasses" with key set to "name" 
 #
  ---------------------------------------------------------------------- */
bool ToglMaster::RegisterWidgetFunc(
		char *name,  // widget class name
		CreateWidget func // create new instance function 
) {
	std::pair<std::string, CreateWidget> p1(name, func);
	std::pair<ClassTable::iterator, bool> p = toglClasses.insert(p1);
	if (p.second) {
		; // OK
		return true;
	} else {
		cerr<<"Error in ToglMaster:RegisterWidgetFunc(char *, CreateWidget); # ";
		cerr<<"the widget class : "<<"\""<<name<<"\" already exists!"<<endl;
		return false;
	}
}

/*----------------------------------------------------------------------
 #
 # ToglMaster command string: The string has to be register with
 # Togl in order to be become qualified as a togl widget command argument. 
 # A command argument can be register only once, however different classes 
 # of ToglWidgets may interpret it in a different way.
 #
  ---------------------------------------------------------------------- */
bool ToglMaster::RegisterCmdStr(
		const char *name // Togl command string
) {
	if (name != NULL) {
		Togl_CreateCommand((char*)name, ToglMaster::Togl_command);
		return true;
	}
	return false;
}

/*----------------------------------------------------------------------
 #
 # ToglMaster command strings: Each string has to be register with
 # Togl in order to be become qualified as a togl widget command argument.
 # A command argument can be register only once, however different classes 
 # of ToglWidgets may interpret it in a different way.
 #
  ---------------------------------------------------------------------- */
bool ToglMaster::RegisterCmdStrings(
		char *name, ... // list of NULL terminated Togl command srtings
) {
	if (name !=NULL) {
		char *i = name;
		va_list marker;
		va_start(marker, name); // Initialize variable arguments.
		while(i != NULL) {
			Togl_CreateCommand(i, ToglMaster::Togl_command);
			i = va_arg(marker, char*);
		}
		va_end( marker ); // Reset variable arguments
		return true;
	}
	return false;
}

/*----------------------------------------------------------------------
 #
 # ToglMaster registers Togl callbacks 
 #
  ---------------------------------------------------------------------- */
void ToglMaster::InitTogl() {
	Togl_CreateFunc(ToglMaster::Init);
	Togl_DisplayFunc(ToglMaster::Display);
	Togl_ReshapeFunc(ToglMaster::Reshape);
	Togl_DestroyFunc(ToglMaster::Destroy);
	Togl_TimerFunc(ToglMaster::Timer);
	Togl_OverlayDisplayFunc(ToglMaster::Overlay);
}

