//----------------------------------------------
//
// C++ interface "ToglMaster", for "Togl"
// Anargyros Sarafopoulos 04/June/01
// asarafop@bournemouth.ac.uk
//
//---------------------------------------------

#ifndef __TOGL_MASTER_H__
#define __TOGL_MASTER_H__

#if defined(WIN32)
#	if defined(_MSC_VER)
#		pragma warning(disable: 4786)
#	endif
#endif

extern "C" {
#include "togl.h"
}

#include <iostream>
#include <string>
#include <map>

/*
 * Togl 'C' struct
 */
typedef struct Togl Togl;


/*
 * forward declarations
 */
class ToglCmd;
class ToglWidget;
class ToglMaster;

/* 
 * pointer to static Java like "newInstance" method
 */
typedef ToglWidget* (*CreateWidget) ();

/*
 * associative arrays 
 */
typedef std::map<std::string, CreateWidget> ClassTable;
typedef std::map<std::string, ToglWidget*> WidgetTable;
typedef std::multimap<std::string, ToglCmd*> CommandTable;

/*
 * abstract class respresenting a Togl command
 */
class ToglCmd {
	std::string name; // command name
public:
	// exception class
	class WrongArgs {
		std::string message;
	public :
		WrongArgs(const std::string &str) { message = str; }
		WrongArgs(char *str) { message = str; }
		const char* GetMsg() { return message.c_str(); }
	}; 

	// defaut constructor
	ToglCmd() { ; }
	// constructor 
	ToglCmd(std::string str, ToglWidget &w); 
	// virtual destructor
	virtual ~ToglCmd() { ; }
	// Get command name 
	const char* GetName() { return name.c_str(); }
	// interface
	virtual bool Cmd(int argc, const char *argv[]) =0;
};

/* 
 * abstract class representing a (Togl) widget
 */
class ToglWidget {
	Togl *togl; // pointer to Togl 'C' struct
	std::string id;  // identifying string for this widget
	CommandTable commands; // table of widget commands 

	/*
	 * called by ToglMaster to insert "this" widget
	 * into the ToglMaster::toglWidgets table
	 */
	void RegisterWidget(Togl *togl, std::string &str); 

	/*
	 * Called by a ToglCmd object in order to insert
	 * itself (or another ToglCmd) in the ToglWidget::commands table
	 */
	void RegisterCommand(std::string cmdname, ToglCmd* cmd); 

	// allows inserting widgets in "ToglMaster::toglWidgets" table
	friend class ToglMaster; 
	// allows commands to be inserted in "ToglWidget::commands" table
	friend class ToglCmd;
public:

	/*
	 * Rendering-related Commands
	 */

	/*
	 * Signals that the widget should be redrawn. When Tk
	 * is next idle the user's render callback will be
	 * invoked. 
	 */
	void PostRedisplay() { Togl_PostRedisplay(togl); }

	/* 
	 * Swaps the front and back color buffers for a 
	 * double-buffered widget.
	 */
	void SwapBuffers() { Togl_SwapBuffers(togl); }

	/*
	 * Bind the OpenGL rendering context to the specified 
	 * Togl widget.
	 */
	void MakeCurrent() { Togl_MakeCurrent(togl); }
	
	/*
	 * Query Functions
	 */

	/* 
	 * Return the width or height of the given Togl widget. 
	 * Queries typically called within the ReshapeFunc().
	 */
	int GetWidth() { return Togl_Width(togl); }
	int GetHeight() { return Togl_Height(togl); }

	/*
	 * Returns a pointer to the identification string 
	 * associated with a Togl widget or NULL if there's 
	 * no identifier string. 
	 */
	char* GetIdent() { return Togl_Ident(togl); }

	/*
	 * Returns the Tcl interpreter associated with the 
	 * given Togl widget
	 */
	Tcl_Interp *GetInterp() { return Togl_Interp(togl); }

	/*
	 * Color Index Mode Functions
	 */

	/* 
	 * Allocate a color from a read-only colormap. Given a color 
	 * specified by red, green, and blue return a colormap index 
	 * (aka pixel value) whose entry most closely matches the red, 
	 * green, blue color. Red, green, and blue are values in [0,1]. 
	 * This function is only used in color index mode when the 
	 * -privatecmap option is false. 
	 */
	unsigned long AllocColor(float red, float green, float blue) {
		return Togl_AllocColor(togl, red, green, blue);
	}

	/*
	 * Free a color in a read-only colormap. Index is a value which 
	 * was returned by the Togl_AllocColor() function. This function 
	 * is only used in color index mode when the -privatecmap option 
	 * is false.
	 */
	void FreeColor(unsigned long index) {
		Togl_FreeColor(togl, index);
	}

	/* 
	 * Load the colormap entry specified by index with the given 
	 * red, green and blue values. Red, green, and blue are values 
	 * in [0,1]. This function is only used in color index mode when
	 * the -privatecmap option is true. 
	 */
	void SetColor(int index, float red, float green, float blue) {
		Togl_SetColor(togl, index, red, green, blue); 
	}

	/*
	 * Font Functions
	 */

	/*
	 * GLuint Togl_LoadBitmapFont( struct Togl *togl, const char *fontname ) 
	 * Load the named font as a set of glBitmap display lists. fontname may be one of 
	 * TOGL_BITMAP_8_BY_13 
	 * TOGL_BITMAP_9_BY_15 
	 * TOGL_BITMAP_TIMES_ROMAN_10 
	 * TOGL_BITMAP_TIMES_ROMAN_24 
	 * TOGL_BITMAP_HELVETICA_10 
	 * TOGL_BITMAP_HELVETICA_12 
	 * TOGL_BITMAP_HELVETICA_18 
	 * or any X11 font name 
	 * Zero is returned if this function fails. 
	 * After Togl_LoadBitmapFont() has been called, returning fontbase, you can 
	 * render a string s with:
	 * 
	 * glListBase( fontbase ); 
	 * glCallLists( strlen(s), GL_BYTE, s ); 
	 * 
	 * To maximize the portability of your application it is best to use one of the 
	 * predefined TOGL_BITMAP_* fonts.
	 */
	GLuint LoadBitmapFont(const char *fontname) {
		return Togl_LoadBitmapFont(togl, fontname);
	}
	
	/* 
	 * Destroys the bitmap display lists created by by 
	 * ToglWidget::LoadBitmapFont(const char*) 
	 */
	void UnloadBitmapFont(GLuint fontbase) {
		Togl_UnloadBitmapFont(togl, fontbase);
	}

	/*
	 * Overlay Functions
	 */

	/* 
	 * Select the layer into which subsequent OpenGL 
	 * rendering will be directed. layer may be either 
	 * TOGL_OVERLAY or TOGL_NORMAL. 
	 */
	void UseLayer(int layer) {
		Togl_UseLayer(togl, layer);
	}

	// Display the overlay planes, if any. 
	void ShowOverlay() {
		Togl_ShowOverlay(togl);
	}

	// Hide the overlay planes, if any.
	void HideOverlay() {
		Togl_HideOverlay(togl);
	}

	/* 
	 * Signal that the overlay planes should be redraw. When Tk is next 
	 * idle the user's C overlay display callback will be invoked. This is 
	 * typically called from within a Togl sub-command which was registered 
	 * with Togl_CreateCommand(). 
	 */
	void PostOverlayRedisplay() {
		Togl_PostOverlayRedisplay(togl);
	}

	// Returns 1 if overlay planes exist, 0 otherwise. 
	int ExistsOverlay() {
		return Togl_ExistsOverlay(togl);
	}
		
	/* 
	 * Returns the color index of the overlay's 
	 * transparent pixel value.
	 */
	int GetOverlayTransparentValue() {
		return Togl_GetOverlayTransparentValue(togl);
	}

	/*
	 * Returns 1 if the overlay planes are currently 
	 * displayed, 0 otherwise.
	 */
	int IsMappedOverlay() {
		return Togl_IsMappedOverlay(togl);
	}

	/*
	 * Allocate a color in the overlay planes. Red, green, and 
	 * blue are values in [0,1]. Return the color index or -1 
	 * if the allocation fails. 
	 */
	unsigned long AllocColorOverlay(float red, float green, float blue) {
		return Togl_AllocColorOverlay(togl, red, green, blue);
	}

	/* 
	 * Free a color which was allocated with 
	 * ToglWidget::AllocColorOverlay(float, float, float)
	 */
	void FreeColorOverlay(unsigned long index) {
		Togl_FreeColorOverlay(togl, index);
	}

	/*
	 * X11-only Functions
	 * These functions are only implemented on systems 
	 * using the X Window System. We recommend that you avoid 
	 * using these functions in your application since they 
	 * are not portable to other operating/window systems. 
    * If you choose to use these functions in your 
	 * application your must first define the TOGL_X11 
	 * preprocessor symbol before including the Togl header file:
	 * #define TOGL_X11
    * #include "Togl.h" 
	 */
#if defined(TOGL_X11)
   //Returns the X Display of a Togl widget. 
	Display *GetX11Display() { return Togl_Display(togl); }

	// Returns the X Screen of a Togl widget. 
	Screen *GetX11Screen() { return Togl_Screen(togl); }

	// Returns the X screen number of a Togl widget. 
	int GetX11ScreenNumber() { return Togl_ScreenNumber(togl); }

	// Returns the X Colormap used by a Togl widget. 
	Colormap GetX11Colormap() { return Togl_Colormap(togl); }
#endif // TOGL_X11

	// constructor
	ToglWidget() { togl=NULL; }

	// virtual destructor
	virtual ~ToglWidget() { ; }

	// invoke a widget command
	virtual bool Cmd(int argc, const char *argv[]); 

	// interface 
	virtual void Init()   =0;// initialize
	virtual void Reshape()=0;// viewport update
	virtual void Display()=0;// redraw
	virtual void Destroy()=0;// cleanup OpenGL state
	virtual void Timer()  =0;// timer callback
   virtual void Overlay()=0;// redraw overlay
};

/*
 * ToglMaster interface deals with the 'C' Togl source
 */
class ToglMaster {
	// Table of widget classes 
	static ClassTable toglClasses;
	// Table of widget instances
	static WidgetTable toglWidgets;

	// register Togl widgets

	static bool RegisterToglWidget(const char *wname, ToglWidget &widget);
	static bool RegisterCmdStr(const char *name);

	// Togl callbacks

	static void Init(Togl *togl);
	static void Reshape(Togl *togl);
	static void Display(Togl *togl);
	static void Destroy(Togl *togl);
	static void Timer(Togl *togl);
	static void Overlay(Togl *togl);
	static int Togl_command(Togl *, int argc, const char *argv[]);

	// allows to register widgets in the toglWidgets table
	friend class ToglWidget;
public:
	// Master Constructor
	ToglMaster() { ; }
	// Master Register 
	static bool RegisterWidgetFunc(char *name, CreateWidget func);
	static bool RegisterCmdStrings(char *name, ...);
	// Togl initialize 
	static void InitTogl();
	
};

typedef ToglWidget tkgl; // ...

#endif // __TOGL_MASTER_H__

